
# class Extra:
#     loading = False
#
# def load_extra():
#     if not Extra.loading:
#         Extra.loading = True
# #         from . import useful_contracts
#         from .library import miscellaneous_aliases
#         # And after everything else is loaded, load the  utils
#     else:
#         print('already loading...')

def new_contract_impl(identifier, condition):
    from .._compat import basestring
    from .syntax import ParseException
    from .library.extensions import CheckCallableWithSelf
    from .library import (CheckCallable, Extension, SeparateContext,
        identifier_expression)
    from .interface import describe_value, Where, ContractSyntaxError
    from .inspection import can_accept_self, can_be_used_as_a_type, can_accept_at_least_one_argument

    # Be friendly
    if not isinstance(identifier, basestring):
        msg = 'I expect the identifier to be a string; received %s.' % describe_value(identifier)
        raise ValueError(msg)

    # Make sure it is not already an expression that we know.
    # (exception: allow redundant definitions. To this purpose,
    #   skip this test if the identifier is already known, and catch
    #   later if the condition changed.)
    if identifier in Extension.registrar:
        # already known as identifier; check later if the condition
        # remained the same.
        pass
    else:
        # check it does not redefine list, tuple, etc.
        try:
            c = parse_contract_string_actual(identifier)
            msg = ('Invalid identifier %r; it overwrites an already known '
                   'expression. In fact, I can parse it as %s (%r).' %
                   (identifier, c, c))
            raise ValueError(msg)
        except ContractSyntaxError:
            pass

    # Make sure it corresponds to our idea of identifier
    try:
        c = identifier_expression.parseString(identifier, parseAll=True)
    except ParseException as e:
        loc = e.loc
        if loc >= len(identifier):
            loc -= 1
        where = Where(identifier, character=loc)  #line=e.lineno, column=e.col)
        # msg = 'Error in parsing string: %s' % e
        msg = ('The given identifier %r does not correspond to my idea '
               'of what an identifier should look like;\n%s\n%s'
               % (identifier, e, where))
        raise ValueError(msg)

    # Now let's check the condition
    if isinstance(condition, basestring):
        # We assume it is a condition that should parse cleanly
        try:
            # could call parse_flexible_spec as well here
            bare_contract = parse_contract_string_actual(condition)
        except ContractSyntaxError as e:
            msg = ('The given condition %r does not parse cleanly: %s' %
                   (condition, e))
            raise ValueError(msg)
    # Important: types are callable, so check this first.
    elif can_be_used_as_a_type(condition):
        from .main import parse_flexible_spec
        # parse_flexible_spec can take care of types
        bare_contract = parse_flexible_spec(condition)
    # Lastly, it should be a callable
    elif hasattr(condition, '__call__'):
        # Check that the signature is right
        if can_accept_self(condition):
            bare_contract = CheckCallableWithSelf(condition)
        elif can_accept_at_least_one_argument(condition):
            bare_contract = CheckCallable(condition)
        else:
            raise ValueError("The given callable %r should be able to accept "
                             "at least one argument" % condition)
    else:
        raise ValueError('I need either a string or a callable for the '
                         'condition; found %s.' % describe_value(condition))

    # Separate the context if needed
    if isinstance(bare_contract, (CheckCallable, CheckCallableWithSelf)):
        contract = bare_contract
    else:
        contract = SeparateContext(bare_contract)

    # It's okay if we define the same thing twice
    if identifier in Extension.registrar:
        old = Extension.registrar[identifier]
        if not (contract == old):
            msg = ('Tried to redefine %r with a definition that looks '
                   'different to me.\n' % identifier)
            msg += ' - old: %r\n' % old
            msg += ' - new: %r\n' % contract
            raise ValueError(msg)
    else:
        Extension.registrar[identifier] = contract

    # Before, we check that we can parse it now
    # - not anymore, because since there are possible args/kwargs,
    # - it might be that the keyword alone is not a valid contract
    # if False:
    #     try:
    #         c = parse_contract_string(identifier)
    #         expected = Extension(identifier)
    #         assert c == expected, \
    #             'Expected %r, got %r.' % (c, expected)  # pragma: no cover
    #     except ContractSyntaxError:  # pragma: no cover
    #         #assert False, 'Cannot parse %r: %s' % (identifier, e)
    #         raise

    return contract


def parse_contract_string_actual(string):
    from .interface import (Contract, ContractDefinitionError, ContractSyntaxError,
        Where)
    from .main import Storage, _cacheable, check_param_is_string
    from .syntax import ParseException, ParseFatalException, contract_expression

    check_param_is_string(string)

    if string in Storage.string2contract:
        return Storage.string2contract[string]
    # TODO: This needs a total rewrite to avoid exception handling
    try:
        c = contract_expression.parseString(string,
                                            parseAll=True)[0]
        assert isinstance(c, Contract), 'Want Contract, not %r' % c
        if _cacheable(string, c):
            Storage.string2contract[string] = c
        return c
    except ContractDefinitionError as e:
        raise
    except ParseException as e:
        where = Where(string, character=e.loc)
        msg = '%s' % e
        raise ContractSyntaxError(msg, where=where)
    except ParseFatalException as e:
        where = Where(string, character=e.loc)
        msg = '%s' % e
        raise ContractSyntaxError(msg, where=where)





import dataclasses
import warnings
import os


@dataclasses.dataclass(frozen=True)
class Switches:
    # ALWAYS ON
    disable_all = False


def all_disabled():
    """ Returns true if all contracts checks are disabled. """
    return Switches.disable_all




from Aspidites._vendor.RestrictedPython import compile_restricted_exec
from Aspidites._vendor._compat import IS_PY3
from Aspidites._vendor.RestrictedPython.Guards import guarded_unpack_sequence
from Aspidites.tests._vendor.RestrictedPython.helper import restricted_exec

import pytest


TRY_EXCEPT = """
def try_except(m):
    try:
        m('try')
        raise IndentationError('f1')
    except IndentationError as error:
        m('except')
"""


def test_RestrictingNodeTransformer__visit_Try__1(
        mocker):
    """It allows try-except statements."""
    trace = mocker.stub()
    restricted_exec(TRY_EXCEPT)['try_except'](trace)

    trace.assert_has_calls([
        mocker.call('try'),
        mocker.call('except')
    ])


TRY_EXCEPT_ELSE = """
def try_except_else(m):
    try:
        m('try')
    except:
        m('except')
    else:
        m('else')
"""


def test_RestrictingNodeTransformer__visit_Try__2(
        mocker):
    """It allows try-except-else statements."""
    trace = mocker.stub()
    restricted_exec(TRY_EXCEPT_ELSE)['try_except_else'](trace)

    trace.assert_has_calls([
        mocker.call('try'),
        mocker.call('else')
    ])


TRY_FINALLY = """
def try_finally(m):
    try:
        m('try')
        1 / 0
    finally:
        m('finally')
        return
"""


def test_RestrictingNodeTransformer__visit_TryFinally__1(
        mocker):
    """It allows try-finally statements."""
    trace = mocker.stub()
    restricted_exec(TRY_FINALLY)['try_finally'](trace)

    trace.assert_has_calls([
        mocker.call('try'),
        mocker.call('finally')
    ])


TRY_EXCEPT_FINALLY = """
def try_except_finally(m):
    try:
        m('try')
        1 / 0
    except:
        m('except')
    finally:
        m('finally')
"""


def test_RestrictingNodeTransformer__visit_TryFinally__2(
        mocker):
    """It allows try-except-finally statements."""
    trace = mocker.stub()
    restricted_exec(TRY_EXCEPT_FINALLY)['try_except_finally'](trace)

    trace.assert_has_calls([
        mocker.call('try'),
        mocker.call('except'),
        mocker.call('finally')
    ])


TRY_EXCEPT_ELSE_FINALLY = """
def try_except_else_finally(m):
    try:
        m('try')
    except:
        m('except')
    else:
        m('else')
    finally:
        m('finally')
"""


def test_RestrictingNodeTransformer__visit_TryFinally__3(
        mocker):
    """It allows try-except-else-finally statements."""
    trace = mocker.stub()
    restricted_exec(TRY_EXCEPT_ELSE_FINALLY)['try_except_else_finally'](trace)

    trace.assert_has_calls([
        mocker.call('try'),
        mocker.call('else'),
        mocker.call('finally')
    ])


EXCEPT_WITH_TUPLE_UNPACK = """
def tuple_unpack(err):
    try:
        raise err
    except Exception as (a, (b, c)):
        return a + b + c
"""


@pytest.mark.skipif(
    IS_PY3,
    reason="tuple unpacking on exceptions is gone in python3")
def test_RestrictingNodeTransformer__visit_ExceptHandler__1(
        mocker):  # pragma: no cover
    _getiter_ = mocker.stub()
    _getiter_.side_effect = lambda it: it

    glb = {
        '_getiter_': _getiter_,
        '_unpack_sequence_': guarded_unpack_sequence
    }

    restricted_exec(EXCEPT_WITH_TUPLE_UNPACK, glb)
    err = Exception(1, (2, 3))
    ret = glb['tuple_unpack'](err)
    assert ret == 6

    _getiter_.assert_has_calls([
        mocker.call(err),
        mocker.call((2, 3))])


BAD_TRY_EXCEPT = """
def except_using_bad_name():
    try:
        foo
    except NameError as _leading_underscore:
        # The name of choice (say, _write) is now assigned to an exception
        # object.  Hard to exploit, but conceivable.
        pass
"""


def test_RestrictingNodeTransformer__visit_ExceptHandler__2():
    """It denies bad names in the except as statement."""
    result = compile_restricted_exec(BAD_TRY_EXCEPT)
    assert result.errors == (
        'Line 5: "_leading_underscore" is an invalid variable name because '
        'it starts with "_"',)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) The python-semanticversion project
# This code is distributed under the two-clause BSD License.

"""Test the various functions from 'base'."""

import sys
import unittest

import pytest

from Aspidites._vendor.semantic_version import (
    Spec,
    SpecItem,
    Version,
    compare,
    match,
    validate,
)

pytestmark = pytest.mark.filterwarnings(
    "ignore::DeprecationWarning", "ignore::PendingDeprecationWarning"
)


class TopLevelTestCase(unittest.TestCase):
    """Test module-level functions."""

    if sys.version_info[0] <= 2:
        import contextlib

        @contextlib.contextmanager
        def subTest(self, **kwargs):
            yield

    versions = (
        ("0.1.0", "0.1.1", -1),
        ("0.1.1", "0.1.1", 0),
        ("0.1.1", "0.1.0", 1),
        ("0.1.0-alpha", "0.1.0", -1),
        ("0.1.0-alpha+2", "0.1.0-alpha", NotImplemented),
    )

    def test_compare(self):
        for a, b, expected in self.versions:
            with self.subTest(a=a, b=b):
                result = compare(a, b)
                self.assertEqual(
                    expected,
                    result,
                    "compare(%r, %r) should be %r instead of %r"
                    % (a, b, expected, result),
                )

    matches = (
        (">=0.1.1", "0.1.2"),
        (">=0.1.1", "0.1.1"),
        (">=0.1.1", "0.1.2-alpha"),
        (">=0.1.1,!=0.2.0", "0.2.1"),
    )

    def test_match(self):
        for spec, version in self.matches:
            with self.subTest(spec=spec, version=version):
                self.assertTrue(
                    match(spec, version), "%r should accept %r" % (spec, version)
                )

    valid_strings = (
        "1.0.0-alpha",
        "1.0.0-alpha.1",
        "1.0.0-beta.2",
        "1.0.0-beta.11",
        "1.0.0-rc.1",
        "1.0.0-rc.1+build.1",
        "1.0.0",
        "1.0.0+0.3.7",
        "1.3.7+build",
        "1.3.7+build.2.b8f12d7",
        "1.3.7+build.11.e0f985a",
        "1.1.1",
        "1.1.2",
        "1.1.3-rc4.5",
        "1.1.3-rc42.3-14-15.24+build.2012-04-13.223",
        "1.1.3+build.2012-04-13.HUY.alpha-12.1",
    )

    def test_validate_valid(self):
        for version in self.valid_strings:
            with self.subTest(version=version):
                self.assertTrue(
                    validate(version), "%r should be a valid version" % (version,)
                )

    invalid_strings = (
        "1",
        "v1",
        "1.2.3.4",
        "1.2",
        "1.2a3",
        "1.2.3a4",
        "v12.34.5",
        "1.2.3+4+5",
    )

    def test_validate_invalid(self):
        for version in self.invalid_strings:
            with self.subTest(version=version):
                self.assertFalse(
                    validate(version), "%r should not be a valid version" % (version,)
                )


class VersionTestCase(unittest.TestCase):
    if sys.version_info[0] <= 2:
        import contextlib

        @contextlib.contextmanager
        def subTest(self, **kwargs):
            yield

    versions = {
        "1.0.0-alpha": (1, 0, 0, ("alpha",), ()),
        "1.0.0-alpha.1": (1, 0, 0, ("alpha", "1"), ()),
        "1.0.0-beta.2": (1, 0, 0, ("beta", "2"), ()),
        "1.0.0-beta.11": (1, 0, 0, ("beta", "11"), ()),
        "1.0.0-rc.1": (1, 0, 0, ("rc", "1"), ()),
        "1.0.0-rc.1+build.1": (1, 0, 0, ("rc", "1"), ("build", "1")),
        "1.0.0": (1, 0, 0, (), ()),
        "1.0.0+0.3.7": (1, 0, 0, (), ("0", "3", "7")),
        "1.3.7+build": (1, 3, 7, (), ("build",)),
        "1.3.7+build.2.b8f12d7": (1, 3, 7, (), ("build", "2", "b8f12d7")),
        "1.3.7+build.11.e0f985a": (1, 3, 7, (), ("build", "11", "e0f985a")),
        "1.1.1": (1, 1, 1, (), ()),
        "1.1.2": (1, 1, 2, (), ()),
        "1.1.3-rc4.5": (1, 1, 3, ("rc4", "5"), ()),
        "1.1.3-rc42.3-14-15.24+build.2012-04-13.223": (
            1,
            1,
            3,
            ("rc42", "3-14-15", "24"),
            ("build", "2012-04-13", "223"),
        ),
        "1.1.3+build.2012-04-13.HUY.alpha-12.1": (
            1,
            1,
            3,
            (),
            ("build", "2012-04-13", "HUY", "alpha-12", "1"),
        ),
    }

    def test_parsing(self):
        for text, expected_fields in self.versions.items():
            with self.subTest(text=text):
                version = Version(text)
                actual_fields = (
                    version.major,
                    version.minor,
                    version.patch,
                    version.prerelease,
                    version.build,
                )
                self.assertEqual(expected_fields, actual_fields)

    def test_str(self):
        for text in self.versions:
            with self.subTest(text=text):
                version = Version(text)
                self.assertEqual(text, str(version))
                self.assertEqual("Version('%s')" % text, repr(version))

    def test_compare_to_self(self):
        for text in self.versions:
            with self.subTest(text=text):
                self.assertEqual(Version(text), Version(text))
                self.assertNotEqual(text, Version(text))

    partial_versions = {
        "1.1": (1, 1, None, None, None),
        "2": (2, None, None, None, None),
        "1.0.0-alpha": (1, 0, 0, ("alpha",), None),
        "1.0.0-alpha.1": (1, 0, 0, ("alpha", "1"), None),
        "1.0.0-beta.2": (1, 0, 0, ("beta", "2"), None),
        "1.0.0-beta.11": (1, 0, 0, ("beta", "11"), None),
        "1.0.0-rc.1": (1, 0, 0, ("rc", "1"), None),
        "1.0.0": (1, 0, 0, None, None),
        "1.1.1": (1, 1, 1, None, None),
        "1.1.2": (1, 1, 2, None, None),
        "1.1.3-rc4.5": (1, 1, 3, ("rc4", "5"), None),
        "1.0.0-": (1, 0, 0, (), None),
        "1.0.0-rc.1+build.1": (1, 0, 0, ("rc", "1"), ("build", "1")),
        "1.0.0+0.3.7": (1, 0, 0, (), ("0", "3", "7")),
        "1.3.7+build": (1, 3, 7, (), ("build",)),
        "1.3.7+build.2.b8f12d7": (1, 3, 7, (), ("build", "2", "b8f12d7")),
        "1.3.7+build.11.e0f985a": (1, 3, 7, (), ("build", "11", "e0f985a")),
        "1.1.3-rc42.3-14-15.24+build.2012-04-13.223": (
            1,
            1,
            3,
            ("rc42", "3-14-15", "24"),
            ("build", "2012-04-13", "223"),
        ),
        "1.1.3+build.2012-04-13.HUY.alpha-12.1": (
            1,
            1,
            3,
            (),
            ("build", "2012-04-13", "HUY", "alpha-12", "1"),
        ),
    }

    def test_parsing_partials(self):
        for text, expected_fields in self.partial_versions.items():
            with self.subTest(text=text):
                version = Version(text, partial=True)
                actual_fields = (
                    version.major,
                    version.minor,
                    version.patch,
                    version.prerelease,
                    version.build,
                )
                self.assertEqual(expected_fields, actual_fields)
                self.assertTrue(
                    version.partial, "%r should have partial=True" % version
                )

    def test_str_partials(self):
        for text in self.partial_versions:
            with self.subTest(text=text):
                version = Version(text, partial=True)
                self.assertEqual(text, str(version))
                self.assertEqual("Version('%s', partial=True)" % text, repr(version))

    def test_compare_partial_to_self(self):
        for text in self.partial_versions:
            with self.subTest(text=text):
                self.assertEqual(
                    Version(text, partial=True), Version(text, partial=True)
                )
                self.assertNotEqual(text, Version(text, partial=True))

    def test_hash(self):
        self.assertEqual(1, len(set([Version("0.1.0"), Version("0.1.0")])))

        self.assertEqual(
            2, len(set([Version("0.1.0"), Version("0.1.0", partial=True)]))
        )

        # A fully-defined 'partial' version isn't actually partial.
        self.assertEqual(
            1, len(set([Version("0.1.0-a1+34"), Version("0.1.0-a1+34", partial=True)]))
        )

    @unittest.skipIf(
        sys.version_info[0] <= 2, "Comparisons don't raise TypeError in Python 2"
    )
    def test_invalid_comparisons(self):
        v = Version("0.1.0")
        with self.assertRaises(TypeError):
            v < "0.1.0"
        with self.assertRaises(TypeError):
            v <= "0.1.0"
        with self.assertRaises(TypeError):
            v > "0.1.0"
        with self.assertRaises(TypeError):
            v >= "0.1.0"

        self.assertTrue(v != "0.1.0")
        self.assertFalse(v == "0.1.0")

    def test_bump_clean_versions(self):
        # We Test each property explicitly as the == comparator for versions
        # does not distinguish between prerelease or builds for equality.

        v = Version("1.0.0+build")
        v = v.next_major()
        self.assertEqual(v.major, 2)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.0.0+build")
        v = v.next_minor()
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 1)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.0.0+build")
        v = v.next_patch()
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.patch, 1)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.1.0+build")
        v = v.next_major()
        self.assertEqual(v.major, 2)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.1.0+build")
        v = v.next_minor()
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 2)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.1.0+build")
        v = v.next_patch()
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 1)
        self.assertEqual(v.patch, 1)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.0.1+build")
        v = v.next_major()
        self.assertEqual(v.major, 2)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.0.1+build")
        v = v.next_minor()
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 1)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.0.1+build")
        v = v.next_patch()
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.patch, 2)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

    def test_bump_prerelease_versions(self):
        # We Test each property explicitly as the == comparator for versions
        # does not distinguish between prerelease or builds for equality.

        v = Version("1.0.0-pre+build")
        v = v.next_major()
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.0.0-pre+build")
        v = v.next_minor()
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.0.0-pre+build")
        v = v.next_patch()
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.1.0-pre+build")
        v = v.next_major()
        self.assertEqual(v.major, 2)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.1.0-pre+build")
        v = v.next_minor()
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 1)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.1.0-pre+build")
        v = v.next_patch()
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 1)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.0.1-pre+build")
        v = v.next_major()
        self.assertEqual(v.major, 2)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.0.1-pre+build")
        v = v.next_minor()
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 1)
        self.assertEqual(v.patch, 0)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())

        v = Version("1.0.1-pre+build")
        v = v.next_patch()
        self.assertEqual(v.major, 1)
        self.assertEqual(v.minor, 0)
        self.assertEqual(v.patch, 1)
        self.assertEqual(v.prerelease, ())
        self.assertEqual(v.build, ())


class SpecItemTestCase(unittest.TestCase):
    if sys.version_info[0] <= 2:
        import contextlib

        @contextlib.contextmanager
        def subTest(self, **kwargs):
            yield

    invalids = [
        "<=0.1.1+build3",
        "<=0.1.1+",
        ">0.2.3-rc2+",
    ]

    def test_invalids(self):
        for invalid in self.invalids:
            with self.subTest(invalid=invalid):
                with self.assertRaises(
                    ValueError, msg="SpecItem(%r) should be invalid" % invalid
                ):
                    SpecItem(invalid)

    components = {
        "==0.1.0": (SpecItem.KIND_EQUAL, 0, 1, 0, None, None),
        "==0.1.2-rc3": (SpecItem.KIND_EQUAL, 0, 1, 2, ("rc3",), None),
        "==0.1.2+build3.14": (SpecItem.KIND_EQUAL, 0, 1, 2, (), ("build3", "14")),
        "<=0.1.1": (SpecItem.KIND_LTE, 0, 1, 1, None, None),
        "<0.1.1": (SpecItem.KIND_LT, 0, 1, 1, None, None),
        "!=0.1.1+": (SpecItem.KIND_NEQ, 0, 1, 1, (), ()),
        "<=0.1.1-": (SpecItem.KIND_LTE, 0, 1, 1, (), None),
        ">=0.2.3-rc2": (SpecItem.KIND_GTE, 0, 2, 3, ("rc2",), None),
        ">=2.0.0": (SpecItem.KIND_GTE, 2, 0, 0, None, None),
        "!=0.1.1+rc3": (SpecItem.KIND_NEQ, 0, 1, 1, (), ("rc3",)),
        "!=0.3.0": (SpecItem.KIND_NEQ, 0, 3, 0, None, None),
        "=0.3.0": (SpecItem.KIND_EQUAL, 0, 3, 0, None, None),
        "0.3.0": (SpecItem.KIND_EQUAL, 0, 3, 0, None, None),
        "~0.1.2": (SpecItem.KIND_TILDE, 0, 1, 2, None, None),
        "^0.1.3": (SpecItem.KIND_CARET, 0, 1, 3, None, None),
    }

    def test_components(self):
        for spec_text, components in self.components.items():
            with self.subTest(spec_text=spec_text):
                kind, major, minor, patch, prerelease, build = components
                spec = SpecItem(spec_text)

                self.assertEqual(kind, spec.kind)
                self.assertEqual(major, spec.spec.major)
                self.assertEqual(minor, spec.spec.minor)
                self.assertEqual(patch, spec.spec.patch)
                self.assertEqual(prerelease, spec.spec.prerelease)
                self.assertEqual(build, spec.spec.build)

    matches = {
        "==0.1.0": (
            ["0.1.0", "0.1.0+build1"],
            ["0.0.1", "0.1.0-rc1", "0.2.0", "0.1.1", "0.1.0-rc1+build2"],
        ),
        "=0.1.0": (
            ["0.1.0", "0.1.0+build1"],
            ["0.0.1", "0.1.0-rc1", "0.2.0", "0.1.1", "0.1.0-rc1+build2"],
        ),
        "0.1.0": (
            ["0.1.0", "0.1.0+build1"],
            ["0.0.1", "0.1.0-rc1", "0.2.0", "0.1.1", "0.1.0-rc1+build2"],
        ),
        "==0.1.2-rc3": (
            ["0.1.2-rc3+build1", "0.1.2-rc3+build4.5"],
            ["0.1.2-rc4", "0.1.2", "0.1.3"],
        ),
        "==0.1.2+build3.14": (
            ["0.1.2+build3.14"],
            ["0.1.2-rc+build3.14", "0.1.2+build3.15"],
        ),
        "<=0.1.1": (
            ["0.0.0", "0.1.1-alpha1", "0.1.1", "0.1.1+build2"],
            ["0.1.2"],
        ),
        "<0.1.1": (
            ["0.1.0", "0.0.0"],
            ["0.1.1", "0.1.1-zzz+999", "1.2.0", "0.1.1+build3"],
        ),
        "<=0.1.2": (
            ["0.1.2+build4", "0.1.2-alpha", "0.1.0"],
            ["0.2.3", "1.1.1", "0.1.3"],
        ),
        "<0.1.1-": (
            ["0.1.0", "0.1.1-alpha", "0.1.1-alpha+4"],
            ["0.2.0", "1.0.0", "0.1.1", "0.1.1+build1"],
        ),
        ">=0.2.3-rc2": (
            ["0.2.3-rc3", "0.2.3", "0.2.3+1", "0.2.3-rc2", "0.2.3-rc2+1"],
            ["0.2.3-rc1", "0.2.2"],
        ),
        ">=0.2.3": (
            ["0.2.3", "0.2.3+1"],
            ["0.2.3-rc3", "0.2.3-rc2", "0.2.3-rc2+1", "0.2.3-rc1", "0.2.2"],
        ),
        "==0.2.3+": (
            ["0.2.3"],
            ["0.2.3+rc1", "0.2.4", "0.2.3-rc2"],
        ),
        "!=0.2.3-rc2+12": (
            ["0.2.3-rc3", "0.2.3", "0.2.3-rc2+1", "0.2.4", "0.2.3-rc3+12"],
            ["0.2.3-rc2+12"],
        ),
        "==2.0.0+b1": (
            ["2.0.0+b1"],
            ["2.1.1", "1.9.9", "1.9.9999", "2.0.0", "2.0.0-rc4"],
        ),
        "!=0.1.1": (
            ["0.1.2", "0.1.0", "1.4.2"],
            ["0.1.1", "0.1.1-alpha", "0.1.1+b1"],
        ),
        "!=0.3.4-": (
            ["0.4.0", "1.3.0", "0.3.4-alpha", "0.3.4-alpha+b1"],
            ["0.3.4", "0.3.4+b1"],
        ),
        "~1.1.2": (
            ["1.1.3", "1.1.2+b1"],
            ["1.1.1", "1.1.2-alpha", "1.1.2-alpha+b1", "1.2.1", "2.1.0"],
        ),
        "^1.1.2": (
            ["1.1.3", "1.1.2+b1", "1.2.1"],
            ["1.1.1", "1.1.2-alpha", "1.1.2-alpha+b1", "2.1.0"],
        ),
        "^0.1.2": (
            ["0.1.2", "0.1.2+b1", "0.1.3"],
            ["0.1.2-alpha", "0.2.0", "1.1.2", "0.1.1"],
        ),
        "^0.0.2": (
            ["0.0.2", "0.0.2+abb"],
            ["0.0.2-alpha", "0.1.0", "0.0.3", "1.0.0"],
        ),
        "~=1.4.5": (
            ["1.4.5", "1.4.10-alpha", "1.4.10"],
            ["1.3.6", "1.4.4", "1.4.5-alpha", "1.5.0"],
        ),
        "~=1.4.0": (
            ["1.4.0", "1.4.10-alpha", "1.4.10"],
            ["1.3.6", "1.3.9", "1.4.0-alpha", "1.5.0"],
        ),
        "~=1.4": (
            ["1.4.0", "1.6.10-alpha", "1.6.10"],
            ["1.3.0", "1.4.0-alpha", "2.0.0"],
        ),
    }

    def test_matches(self):
        for spec_text, versions in self.matches.items():
            spec = SpecItem(spec_text)
            matching, failing = versions

            for version_text in matching:
                with self.subTest(spec=spec_text, version=version_text):
                    version = Version(version_text)
                    self.assertTrue(
                        spec.match(version), "%r should match %r" % (version, spec)
                    )

            for version_text in failing:
                with self.subTest(spec=spec_text, excluded=version_text):
                    version = Version(version_text)
                    self.assertFalse(
                        spec.match(version), "%r should not match %r" % (version, spec)
                    )

    def test_equality(self):
        spec1 = SpecItem("==0.1.0")
        spec2 = SpecItem("==0.1.0")
        self.assertEqual(spec1, spec2)
        self.assertFalse(spec1 == "==0.1.0")

    def test_to_string(self):
        spec = SpecItem("==0.1.0")
        self.assertEqual("==0.1.0", str(spec))
        self.assertEqual(SpecItem.KIND_EQUAL, spec.kind)

    def test_hash(self):
        self.assertEqual(1, len(set([SpecItem("==0.1.0"), SpecItem("==0.1.0")])))


class CoerceTestCase(unittest.TestCase):
    if sys.version_info[0] <= 2:
        import contextlib

        @contextlib.contextmanager
        def subTest(self, **kwargs):
            yield

    examples = {
        # Dict of target: [list of equivalents]
        "0.0.0": ("0", "0.0", "0.0.0", "0.0.0+", "0-"),
        "0.1.0": ("0.1", "0.1+", "0.1-", "0.1.0", "0.01.0", "000.0001.0000000000"),
        "0.1.0+2": ("0.1.0+2", "0.1.0.2"),
        "0.1.0+2.3.4": ("0.1.0+2.3.4", "0.1.0+2+3+4", "0.1.0.2+3+4"),
        "0.1.0+2-3.4": ("0.1.0+2-3.4", "0.1.0+2-3+4", "0.1.0.2-3+4", "0.1.0.2_3+4"),
        "0.1.0-a2.3": ("0.1.0-a2.3", "0.1.0a2.3", "0.1.0_a2.3"),
        "0.1.0-a2.3+4.5-6": (
            "0.1.0-a2.3+4.5-6",
            "0.1.0a2.3+4.5-6",
            "0.1.0a2.3+4.5_6",
            "0.1.0a2.3+4+5/6",
        ),
    }

    def test_coerce(self):
        for equivalent, samples in self.examples.items():
            target = Version(equivalent)
            for sample in samples:
                with self.subTest(target=equivalent, sample=sample):
                    v_sample = Version.coerce(sample)
                    self.assertEqual(target, v_sample)

    def test_invalid(self):
        self.assertRaises(ValueError, Version.coerce, "v1")


class SpecTestCase(unittest.TestCase):
    if sys.version_info[0] <= 2:
        import contextlib

        @contextlib.contextmanager
        def subTest(self, **kwargs):
            yield

        def assertCountEqual(self, a, b):
            import collections

            self.assertEqual(
                collections.Counter(a),
                collections.Counter(b),
            )

    examples = {
        ">=0.1.1,<0.1.2": [">=0.1.1", "<0.1.2"],
        ">=0.1.0,!=0.1.3-rc1,<0.1.3": [">=0.1.0", "!=0.1.3-rc1", "<0.1.3"],
        "=0.1.2": ["==0.1.2"],
        ">=0.1.2": [">=0.1.2"],
        "^1.2.3": [">=1.2.3", "<2.0.0"],
        "~=1.2.3": [">=1.2.3", "<1.3.0"],
    }

    def test_parsing(self):
        for spec_list_text, specs in self.examples.items():
            with self.subTest(spec=spec_list_text):
                spec_list = Spec(spec_list_text)

                self.assertEqual(spec_list_text, str(spec_list))
                self.assertNotEqual(spec_list_text, spec_list)
                self.assertCountEqual(specs, [str(spec) for spec in spec_list])

    split_examples = {
        (">=0.1.1", "<0.1.2", "!=0.1.1+build1"): [
            ">=0.1.1",
            "<0.1.2",
            "!=0.1.1+build1",
        ],
        (">=0.1.0", "!=0.1.3-rc1,<0.1.3"): [">=0.1.0", "!=0.1.3-rc1", "<0.1.3"],
    }

    def test_parsing_split(self):
        for spec_list_texts, specs in self.split_examples.items():
            with self.subTest(spec=spec_list_texts):
                spec_list = Spec(*spec_list_texts)

                self.assertEqual(",".join(spec_list_texts), str(spec_list))
                self.assertCountEqual(specs, [str(spec) for spec in spec_list])
                self.assertEqual(spec_list, Spec(",".join(spec_list_texts)))

                for spec_text in specs:
                    self.assertIn(str(SpecItem(spec_text)), repr(spec_list))

    matches = {
        # At least 0.1.1 excluding pre-releases, less than 0.1.2 excluding pre-releases
        ">=0.1.1,<0.1.2": (
            ["0.1.1", "0.1.1+4"],
            ["0.1.1-alpha", "0.1.2-alpha", "0.1.2", "1.3.4"],
        ),
        # 0.1.x
        "==0.1.*": (
            ["0.1.1", "0.1.1+4", "0.1.0", "0.1.99"],
            ["0.1.0-alpha", "0.0.1", "0.2.0"],
        ),
        # 1.x.x
        "==1.*": (
            ["1.1.1", "1.1.0+4", "1.1.0", "1.99.99"],
            ["1.0.0-alpha", "0.1.0", "2.0.0"],
        ),
        # At least 0.1.0 with pre-releases, less than 0.1.4 excluding pre-releases,
        # neither 0.1.3-rc1 nor any build of that version,
        # not 0.1.0+b3 precisely
        ">=0.1.0-,!=0.1.3-rc1,!=0.1.0+b3,<0.1.4": (
            ["0.1.1", "0.1.0+b4", "0.1.2", "0.1.3-rc2"],
            [
                "0.0.1",
                "0.1.0+b3",
                "0.1.4",
                "0.1.4-alpha",
                "0.1.3-rc1+4",
                "0.1.0-alpha",
                "0.2.2",
                "0.1.4-rc1",
            ],
        ),
    }

    def test_matches(self):
        for spec_list_text, versions in self.matches.items():
            spec_list = Spec(spec_list_text)
            matching, failing = versions

            for version_text in matching:
                with self.subTest(spec=spec_list_text, matching=version_text):
                    version = Version(version_text)
                    self.assertTrue(
                        version in spec_list,
                        "%r should be in %r" % (version, spec_list),
                    )
                    self.assertTrue(
                        spec_list.match(version),
                        "%r should match %r" % (version, spec_list),
                    )

            for version_text in failing:
                with self.subTest(spec=spec_list_text, excluded=version_text):
                    version = Version(version_text)
                    self.assertFalse(
                        version in spec_list,
                        "%r should not be in %r" % (version, spec_list),
                    )
                    self.assertFalse(
                        spec_list.match(version),
                        "%r should not match %r" % (version, spec_list),
                    )

    def test_equality(self):
        for spec_list_text in self.examples:
            with self.subTest(spec=spec_list_text):
                slist1 = Spec(spec_list_text)
                slist2 = Spec(spec_list_text)
                self.assertEqual(slist1, slist2)
                self.assertFalse(slist1 == spec_list_text)

    def test_filter_empty(self):
        s = Spec(">=0.1.1")
        res = tuple(s.filter(()))
        self.assertEqual((), res)

    def test_filter_incompatible(self):
        s = Spec(">=0.1.1,!=0.1.4")
        res = tuple(
            s.filter(
                [
                    Version("0.1.0"),
                    Version("0.1.4"),
                    Version("0.1.4-alpha"),
                ]
            )
        )
        self.assertEqual((), res)

    def test_filter_compatible(self):
        s = Spec(">=0.1.1,!=0.1.4,<0.2.0")
        res = tuple(
            s.filter(
                [
                    Version("0.1.0"),
                    Version("0.1.1"),
                    Version("0.1.5"),
                    Version("0.1.4-alpha"),
                    Version("0.1.2"),
                    Version("0.2.0-rc1"),
                    Version("3.14.15"),
                ]
            )
        )

        expected = (
            Version("0.1.1"),
            Version("0.1.5"),
            Version("0.1.2"),
        )

        self.assertEqual(expected, res)

    def test_select_empty(self):
        s = Spec(">=0.1.1")
        self.assertIsNone(s.select(()))

    def test_select_incompatible(self):
        s = Spec(">=0.1.1,!=0.1.4")
        res = s.select(
            [
                Version("0.1.0"),
                Version("0.1.4"),
                Version("0.1.4-alpha"),
            ]
        )
        self.assertIsNone(res)

    def test_select_compatible(self):
        s = Spec(">=0.1.1,!=0.1.4,<0.2.0")
        res = s.select(
            [
                Version("0.1.0"),
                Version("0.1.1"),
                Version("0.1.5"),
                Version("0.1.4-alpha"),
                Version("0.1.2"),
                Version("0.2.0-rc1"),
                Version("3.14.15"),
            ]
        )

        self.assertEqual(Version("0.1.5"), res)

    def test_contains(self):
        self.assertFalse("ii" in Spec(">=0.1.1"))

    def test_hash(self):
        self.assertEqual(1, len(set([Spec(">=0.1.1"), Spec(">=0.1.1")])))


if __name__ == "__main__":  # pragma: no cover
    unittest.main()



##############################################################################
# Modified to work as a lightweight alternative to vendoring six.
# Originally from the package RestrictedPython.
# Copyright (c) 2002 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import platform
import sys

_version = sys.version_info
IS_PY2 = _version.major == 2
IS_PY3 = _version.major == 3
IS_PY34_OR_GREATER = _version.major == 3 and _version.minor >= 4
IS_PY35_OR_GREATER = _version.major == 3 and _version.minor >= 5
IS_PY36_OR_GREATER = _version.major == 3 and _version.minor >= 6
IS_PY37_OR_GREATER = _version.major == 3 and _version.minor >= 7
IS_PY38_OR_GREATER = _version.major == 3 and _version.minor >= 8

if IS_PY2:  # pragma: no cover
    basestring = basestring  # NOQA: F821  # Python 2 only built-in function
else:  # pragma: PY3
    basestring = str

if IS_PY2:  # pragma: no cover
    unicode = unicode # NOQA: F821  # Python 2 only built-in function
else:
    unicode = str

IS_CPYTHON = platform.python_implementation() == 'CPython'



good_examples = []
syntax_fail_examples = []
semantic_fail_examples = []
contract_fail_examples = []


# If exact is True, we are providing a canonical form
# for the expression and we want it back.
def good(a, b, exact=True):
    good_examples.append((a, b, exact))


def semantic_fail(a, b, exact=True):
    semantic_fail_examples.append((a, b, exact))


def syntax_fail(s):
    syntax_fail_examples.append(s)


def fail(a, b, exact=True):
    contract_fail_examples.append((a, b, exact))





from . import syntax_fail, good, fail, semantic_fail
from Aspidites._vendor.contracts.library.sets import ASet

good(repr(ASet.__init__(ASet(), length_contract='3', elements_contract='int')), {1, 2, 3})
good('set[3](int)', {1, 2, 3})
fail('set[3](int)', {1, 2})
semantic_fail('set[X](int)', 1)
syntax_fail('set')


""" Other testing examples """

from Aspidites._vendor.contracts.test_registrar import good, fail

# dummy
good(['*'], [0])
good(['*', '*'], [0, 1])
good(['=0', '=1'], [0, 1])
fail(['=0', '=1'], [0, 2])

# Lists of equal length
good(['list[N]', 'list[N]'], [[4], [3]])
good(['list[N]', 'list[N]'], [[], []])
fail(['list[N]', 'list[N]'], [[], [1]])
good(['list[N]', 'list[N],N>0'], [[1], [3]])
# we can also refer to the other context
good(['list[N]', 'list,N>0'], [[1], [3]])
fail(['list[N]', 'list,N>0'], [[], [3]])

# Lists of different length
good(['list[N]', 'list[M],M!=N'], [[4], [3, 2]])
good(['list[N]', 'list[M],M!=N'], [[4, 3], [3]])
fail(['list[N]', 'list[M],M!=N'], [[3], [3]])

# One list shorter than the other
good(['list[N]', 'list[M],M<N'], [[4, 3], [3]])
good(['list[N]', 'list[M],N>M'], [[4, 3], [3]])
fail(['list[N]', 'list[M],N>M'], [[3], [3]])

# Values of the same type
good(['type(x)', 'type(x)'], [0, 1])
good(['type(x)', 'type(x)'], [0.1, 1.1])
fail(['type(x)', 'type(x)'], [0.1, 1])
good(['type(x)', 'type(y)', 'type(x)|type(y)'], [0, 1.1, 1.2])
good(['type(x)', 'type(y)', 'type(x)|type(y)'], [0, 1.1, 3])
fail(['type(x)', 'type(y)', 'type(x)|type(y)'], [0, 1.1, 'ciao'])

# A list with all elements of the same type
good(['list(type(x))'], [[0, 1, 2]])
fail(['list(type(x))'], [[0, 1.2, 2]])
fail(['list(type(x))'], [[0, None, 2]])
# Lists with the same type of elements
good(['list(type(x))', 'list(type(x))'], [[1, 2], [3, 4]])
fail(['list(type(x))', 'list(type(x))'], [[1, 2], [3.0, 4]])
# Using different variables and then imposing they are equal
good(['list(type(x))', 'list(type(y)),x=y'], [[1, 2], [3, 4]])
fail(['list(type(x))', 'list(type(y)),x=y'], [[1, 2], [3.0, 4]])

# a list with at most two types
good('list(type(x|y))', [1, 2, 3, 4.0])
fail('list(type(x|y))', [1, 2, 'ciao', 4.0])


